import json
import logging
import time
from datetime import datetime, timezone
from http import HTTPStatus
from logging import Logger
from typing import List, Union

import click
import uvicorn
from fastapi import FastAPI, Security
from fastapi.middleware.cors import CORSMiddleware
from fastapi.routing import APIRouter
from pydantic import AnyHttpUrl

from isar.apis.models.models import ControlMissionResponse, StartMissionResponse
from isar.apis.robot_control.robot_controller import RobotController
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.security.authentication import Authenticator
from isar.config.settings import settings
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.telemetry.payloads import StartUpMessagePayload
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class API:
    def __init__(
        self,
        authenticator: Authenticator,
        scheduling_controller: SchedulingController,
        robot_controller: RobotController,
        mqtt_publisher: MqttClientInterface,
        port: int = settings.API_PORT,
    ) -> None:
        self.authenticator: Authenticator = authenticator
        self.scheduling_controller: SchedulingController = scheduling_controller
        self.robot_controller: RobotController = robot_controller
        self.host: str = "0.0.0.0"  # Locking uvicorn to use 0.0.0.0
        self.port: int = port
        self.mqtt_publisher: MqttClientInterface = mqtt_publisher

        self.logger: Logger = logging.getLogger("api")

        self.app: FastAPI = self._create_app()
        self.server = self._setup_server()

    def get_app(self) -> FastAPI:
        return self.app

    def _setup_server(self) -> uvicorn.Server:
        config = uvicorn.Config(
            self.app,
            port=self.port,
            host=self.host,
            reload=False,
            log_config=None,
        )
        return uvicorn.Server(config)

    def wait_for_api_server_ready(self) -> None:
        while not self.server.started:
            time.sleep(0.01)
        self.logger.info("Uvicorn server has been started")
        self._publish_startup_message()

    def _create_app(self) -> FastAPI:
        tags_metadata = [
            {
                "name": "Scheduler",
                "description": "Mission functionality",
            }
        ]
        app = FastAPI(
            openapi_tags=tags_metadata,
            on_startup=[
                self.authenticator.load_config,
                self._log_startup_message,
            ],
            swagger_ui_oauth2_redirect_url="/oauth2-redirect",
            swagger_ui_init_oauth={
                "usePkceWithAuthorizationCodeGrant": True,
                "clientId": settings.AZURE_CLIENT_ID,
            },
        )

        if self.authenticator.should_authenticate():
            backend_cors_origins: List[Union[str, AnyHttpUrl]] = [
                f"http://{self.host}:{self.port}"
            ]

            app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        app.include_router(router=self._create_scheduler_router())

        app.include_router(router=self._create_info_router())

        app.include_router(router=self._create_media_control_router())

        return app

    def _create_scheduler_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Scheduler"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/schedule/start-mission",
            endpoint=self.scheduling_controller.start_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Start the mission provided in JSON format",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.UNPROCESSABLE_ENTITY.value: {
                    "description": "Invalid body - The JSON is incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.BAD_REQUEST.value: {
                    "description": "Bad request - Robot does not have the capabilities to perform the mission",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/return-home",
            endpoint=self.scheduling_controller.return_home,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Start return home mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Return home mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.UNPROCESSABLE_ENTITY.value: {
                    "description": "Invalid body - The JSON is incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/stop-mission",
            endpoint=self.scheduling_controller.stop_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Stop the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully stopped",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.UNPROCESSABLE_ENTITY.value: {
                    "description": "Invalid body - The JSON is incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.BAD_REQUEST.value: {
                    "description": "Bad request - Robot does not have the capabilities to perform the mission",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/pause-mission",
            endpoint=self.scheduling_controller.pause_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Pause the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully paused",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/resume-mission",
            endpoint=self.scheduling_controller.resume_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Resume the currently paused mission - if any",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully resumed",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/lockdown",
            endpoint=self.scheduling_controller.lockdown,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Send the robot to dock and ensure that it stays there",
            responses={
                HTTPStatus.OK.value: {"description": "Robot going to dock"},
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state"
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown"
                },
            },
        )
        router.add_api_route(
            path="/schedule/release-lockdown",
            endpoint=self.scheduling_controller.release_lockdown,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Allow the robot to start missions again",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Robot is able to receive missions again"
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state"
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown"
                },
            },
        )
        router.add_api_route(
            path="/schedule/maintenance-mode",
            endpoint=self.scheduling_controller.set_maintenance_mode,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Place the robot in maintenance mode, where it will not run missions before begin released from maintenance mode",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Robot is in maintenance mode and cannot receive missions"
                },
                HTTPStatus.CONFLICT.value: {"description": "Conflict"},
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error"
                },
            },
        )
        router.add_api_route(
            path="/schedule/release-maintenance-mode",
            endpoint=self.scheduling_controller.release_maintenance_mode,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Allow the robot to start missions again",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Robot is able to receive missions again"
                },
                HTTPStatus.CONFLICT.value: {"description": "Conflict"},
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error"
                },
            },
        )
        router.add_api_route(
            path="/schedule/release-intervention-needed",
            endpoint=self.scheduling_controller.release_intervention_needed,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Release the intervention needed state",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Robot released from intervention needed state"
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state"
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown"
                },
            },
        )

        return router

    def _create_info_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Info"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/info/robot-settings",
            endpoint=self.robot_controller.get_info,
            methods=["GET"],
            dependencies=[authentication_dependency],
            summary="Information about the robot-settings",
        )

        return router

    def _create_media_control_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Media"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/media/media-stream-config",
            endpoint=self.robot_controller.generate_media_config,
            methods=["GET"],
            dependencies=[authentication_dependency],
            summary="Generates a media stream connection config",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Media stream was successfully generated",
                },
                HTTPStatus.NO_CONTENT.value: {
                    "description": "Robot has no media config",
                },
            },
        )

        return router

    def _log_startup_message(self) -> None:
        address_format = "%s://%s:%d/docs"
        message = f"Uvicorn running on {address_format} (Press CTRL+C to quit)"
        protocol = "http"
        color_message = (
            "Uvicorn running on "
            + click.style(address_format, bold=True)
            + " (Press CTRL+C to quit)"
        )
        self.logger.info(
            message,
            protocol,
            self.host,
            self.port,
            extra={"color_message": color_message},
        )

    def _publish_startup_message(self) -> None:
        if not self.mqtt_publisher:
            return

        payload: StartUpMessagePayload = StartUpMessagePayload(
            isar_id=settings.ISAR_ID,
            timestamp=datetime.now(timezone.utc),
        )

        self.logger.info("Publishing startup message to MQTT broker")

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_STARTUP,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.await_next_mission import AwaitNextMission
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.returning_home import ReturningHome
from isar.state_machine.states.stopping_paused_mission import StoppingPausedMission
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason


def test_stopping_paused_mission_fails(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = StoppingPausedMission(
        sync_state_machine, "mission_id"
    )
    stopping_paused_mission_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_paused_mission_state.get_event_handler_by_name("failed_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(error_description="", error_reason=ErrorReason.RobotAPIException)
    )

    assert sync_state_machine.events.mqtt_queue.empty()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Paused


def test_stopping_paused_mission_succeeds(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(90.0)
    sync_state_machine.current_state = StoppingPausedMission(
        sync_state_machine, "mission_id"
    )
    stopping_paused_mission_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_paused_mission_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert sync_state_machine.events.mqtt_queue.empty()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is AwaitNextMission


def test_stopping_paused_mission_succeeds_with_low_battery(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = StoppingPausedMission(
        sync_state_machine, "mission_id"
    )
    stopping_paused_mission_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_paused_mission_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert sync_state_machine.events.mqtt_queue.empty()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturningHome



import logging
from threading import Event, Thread
from typing import Optional

from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    ErrorReason,
    RobotAlreadyHomeException,
    RobotException,
    RobotInfeasibleMissionException,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.robot_interface import RobotInterface


class RobotStartMissionThread(Thread):
    def __init__(
        self,
        robot: RobotInterface,
        signal_thread_quitting: Event,
        mission: Mission,
    ):
        self.logger = logging.getLogger("robot")
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.mission = mission
        self.error_message: Optional[ErrorMessage] = None
        Thread.__init__(self, name="Robot start mission thread")

    def run(self) -> None:
        if self.signal_thread_quitting.wait(0):
            return
        try:
            self.robot.initiate_mission(self.mission)
        except RobotAlreadyHomeException as e:
            self.logger.info(
                "Robot disregarded return to home mission as its already at home. Return home mission will be assumed successful without running."
            )
            self.error_message = ErrorMessage(
                error_reason=e.error_reason,
                error_description=e.error_description,
            )
        except RobotInfeasibleMissionException as e:
            self.logger.error(
                f"Mission is infeasible and cannot be scheduled because: {e.error_description}"
            )
            self.error_message = ErrorMessage(
                error_reason=e.error_reason,
                error_description=e.error_description,
            )
        except RobotException as e:
            self.logger.warning(
                f"Initiating mission failed " f"because: {e.error_description}"
            )
            self.error_message = ErrorMessage(
                error_reason=e.error_reason,
                error_description=e.error_description,
            )
        except Exception as e:
            self.logger.warning(
                f"Initiating mission failed due to unknown exception: {e}"
            )
            self.error_message = ErrorMessage(
                error_reason=ErrorReason.RobotUnknownErrorException,
                error_description=str(e),
            )



import logging
import time
from threading import Event, Thread
from typing import Callable, Iterator, Optional

from isar.config.settings import settings
from isar.services.utilities.mqtt_utilities import (
    publish_mission_status,
    publish_task_status,
)
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    ErrorReason,
    RobotCommunicationException,
    RobotCommunicationTimeoutException,
    RobotException,
    RobotMissionStatusException,
    RobotTaskStatusException,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, TaskStatus
from robot_interface.models.mission.task import TASKS, InspectionTask
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface


def get_next_task(task_iterator: Iterator[TASKS]) -> Optional[TASKS]:
    try:
        return next(task_iterator)
    except StopIteration:
        return None


def is_finished(task_status: TaskStatus) -> bool:
    if (
        task_status == TaskStatus.Successful
        or task_status == TaskStatus.PartiallySuccessful
        or task_status == TaskStatus.Cancelled
        or task_status == TaskStatus.Failed
    ):
        return True
    return False


def should_upload_inspections(task: TASKS) -> bool:
    if settings.UPLOAD_INSPECTIONS_ASYNC:
        return False

    return task.status == TaskStatus.Successful and isinstance(task, InspectionTask)


class RobotMonitorMissionThread(Thread):
    def __init__(
        self,
        request_inspection_upload: Callable[[TASKS], None],
        robot: RobotInterface,
        mqtt_publisher: MqttClientInterface,
        signal_thread_quitting: Event,
        signal_mission_stopped: Event,
        mission: Mission,
    ):
        self.logger = logging.getLogger("robot")
        self.request_inspection_upload: Callable[[TASKS], None] = (
            request_inspection_upload
        )
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.signal_mission_stopped: Event = signal_mission_stopped
        self.mqtt_publisher = mqtt_publisher
        self.mission_id: str = mission.id
        self.tasks = mission.tasks

        self.error_message: Optional[ErrorMessage] = None

        Thread.__init__(self, name="Robot mission monitoring thread")

    def _get_task_status(self, task_id: str) -> TaskStatus:
        task_status: TaskStatus = TaskStatus.NotStarted
        failed_task_error: Optional[ErrorMessage] = None
        request_status_failure_counter: int = 0

        while (
            request_status_failure_counter
            < settings.REQUEST_STATUS_FAILURE_COUNTER_LIMIT
        ):
            if self.signal_thread_quitting.wait(0) or self.signal_mission_stopped.wait(
                0
            ):
                failed_task_error = ErrorMessage(
                    error_reason=ErrorReason.RobotTaskStatusException,
                    error_description="Task status collection was cancelled by monitor thread exit",
                )
                break
            if request_status_failure_counter > 0:
                time.sleep(settings.REQUEST_STATUS_COMMUNICATION_RECONNECT_DELAY)

            try:
                task_status = self.robot.task_status(task_id)
                request_status_failure_counter = 0
            except (
                RobotCommunicationTimeoutException,
                RobotCommunicationException,
            ) as e:
                request_status_failure_counter += 1
                self.logger.error(
                    f"Failed to get task status "
                    f"{request_status_failure_counter} times because: "
                    f"{e.error_description}"
                )

                failed_task_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                continue

            except RobotException as e:
                failed_task_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                break

            except Exception as e:
                failed_task_error = ErrorMessage(
                    error_reason=ErrorReason.RobotUnknownErrorException,
                    error_description=str(e),
                )
                break

            return task_status

        raise RobotTaskStatusException(
            error_description=failed_task_error.error_description
        )

    def _get_mission_status(self, mission_id: str) -> MissionStatus:
        mission_status: MissionStatus = MissionStatus.NotStarted
        failed_mission_error: Optional[ErrorMessage] = None
        request_status_failure_counter: int = 0

        while (
            request_status_failure_counter
            < settings.REQUEST_STATUS_FAILURE_COUNTER_LIMIT
        ):
            if self.signal_thread_quitting.wait(0) or self.signal_mission_stopped.wait(
                0
            ):
                break
            if request_status_failure_counter > 0:
                time.sleep(settings.REQUEST_STATUS_COMMUNICATION_RECONNECT_DELAY)

            try:
                mission_status = self.robot.mission_status(mission_id)
                request_status_failure_counter = 0
            except (
                RobotCommunicationTimeoutException,
                RobotCommunicationException,
            ) as e:
                request_status_failure_counter += 1
                self.logger.error(
                    f"Failed to get task status "
                    f"{request_status_failure_counter} times because: "
                    f"{e.error_description}"
                )

                failed_mission_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                continue

            except RobotException as e:
                failed_mission_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                break

            except Exception as e:
                failed_mission_error = ErrorMessage(
                    error_reason=ErrorReason.RobotUnknownErrorException,
                    error_description=str(e),
                )
                break

            return mission_status

        raise RobotMissionStatusException(
            error_description=(
                failed_mission_error.error_description
                if failed_mission_error
                else "Mission status thread cancelled"
            )
        )

    def _log_task_status(self, task: TASKS) -> None:
        if task.status == TaskStatus.Failed:
            self.logger.warning(
                f"Task: {str(task.id)[:8]} was reported as failed by the robot"
            )
        elif task.status == TaskStatus.Successful:
            self.logger.info(
                f"{type(task).__name__} task: {str(task.id)[:8]} completed"
            )
        else:
            self.logger.info(
                f"Task: {str(task.id)[:8]} was reported as {task.status} by the robot"
            )

    def _get_mission_status_based_on_task_status(self) -> MissionStatus:
        fail_statuses = [
            TaskStatus.Cancelled,
            TaskStatus.Failed,
        ]
        partially_fail_statuses = fail_statuses + [TaskStatus.PartiallySuccessful]

        if len(self.tasks) == 0:
            return MissionStatus.Successful
        elif all(task.status in fail_statuses for task in self.tasks):
            return MissionStatus.Failed
        elif any(task.status in partially_fail_statuses for task in self.tasks):
            return MissionStatus.PartiallySuccessful
        else:
            return MissionStatus.Successful

    def _get_and_handle_task_status(self, current_task: TASKS) -> Optional[TASKS]:
        try:
            new_task_status = self._get_task_status(current_task.id)
        except RobotTaskStatusException as e:
            self.logger.error(
                "Failed to collect task status. Error description: %s",
                e.error_description,
            )
            # Currently we only stop mission monitoring after failing to get mission status
            return current_task

        if current_task.status != new_task_status:
            current_task.status = new_task_status
            self._log_task_status(current_task)
            publish_task_status(self.mqtt_publisher, current_task, self.mission_id)

        if is_finished(new_task_status):
            if should_upload_inspections(current_task):
                self.request_inspection_upload(current_task)
            current_task = get_next_task(self.task_iterator)
            if current_task is not None:
                # This is not required, but does make reporting more responsive
                current_task.status = TaskStatus.InProgress
                self._log_task_status(current_task)
                publish_task_status(self.mqtt_publisher, current_task, self.mission_id)
        return current_task

    def _handle_stopped_mission(self, current_task: TASKS) -> None:
        current_task.status = TaskStatus.Cancelled
        publish_task_status(self.mqtt_publisher, current_task, self.mission_id)
        publish_mission_status(
            self.mqtt_publisher,
            self.mission_id,
            MissionStatus.Cancelled,
            self.error_message,
        )

    def run(self) -> None:

        self.task_iterator: Iterator[TASKS] = iter(self.tasks)
        current_task: Optional[TASKS] = get_next_task(self.task_iterator)
        current_task.status = TaskStatus.NotStarted
        current_mission_status = MissionStatus.NotStarted

        while True:
            if self.signal_thread_quitting.wait(0):
                return
            if self.signal_mission_stopped.wait(0):
                self._handle_stopped_mission(current_task)
                return

            if current_task:
                current_task = self._get_and_handle_task_status(current_task)

            new_mission_status: MissionStatus
            try:
                new_mission_status = self._get_mission_status(self.mission_id)
            except RobotMissionStatusException as e:
                self.logger.exception("Failed to collect mission status")
                self.error_message = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                current_task.status = TaskStatus.Failed
                publish_task_status(self.mqtt_publisher, current_task, self.mission_id)
                publish_mission_status(
                    self.mqtt_publisher,
                    self.mission_id,
                    MissionStatus.Failed,
                    self.error_message,
                )
                break

            if new_mission_status == MissionStatus.Cancelled or (
                new_mission_status
                not in [MissionStatus.NotStarted, MissionStatus.InProgress]
                and current_task is None  # We wait for all task statuses
            ):
                # Standardises final mission status report
                new_mission_status = self._get_mission_status_based_on_task_status()
                if self.error_message is None and new_mission_status in [
                    MissionStatus.Failed,
                    MissionStatus.Cancelled,
                ]:
                    self.error_message = ErrorMessage(
                        error_reason=None,
                        error_description="The mission failed because all tasks in the mission failed",
                    )
                publish_mission_status(
                    self.mqtt_publisher,
                    self.mission_id,
                    new_mission_status,
                    self.error_message,
                )
                break

            if new_mission_status != current_mission_status:
                current_mission_status = new_mission_status
                publish_mission_status(
                    self.mqtt_publisher,
                    self.mission_id,
                    current_mission_status,
                    self.error_message,
                )

            time.sleep(settings.FSM_SLEEP_TIME)

        self.logger.info("Stopped monitoring mission")



from enum import Enum


# Did you write your own isar-robot package and would like to have it included here?
# Open a pull request to the ISAR repository!
class RobotModel(Enum):
    TaurobInspector = "TaurobInspector"
    TaurobOperator = "TaurobOperator"
    ExR2 = "ExR2"
    Robot = "Robot"  # This corresponds to the mock in isar_robot
    Turtlebot = "Turtlebot"
    AnymalX = "AnymalX"
    AnymalD = "AnymalD"



from typing import TYPE_CHECKING, List, Union

import isar.state_machine.states.maintenance as Maintenance
import isar.state_machine.states.monitor as Monitor
import isar.state_machine.states.returning_home as ReturningHome
from isar.apis.models.models import MaintenanceResponse
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class StoppingDueToMaintenance(State):

    def __init__(self, state_machine: "StateMachine", mission_id: str):
        events = state_machine.events

        def _failed_stop_event_handler(
            error_message: ErrorMessage,
        ) -> Union[
            Transition[ReturningHome.ReturningHome], Transition[Monitor.Monitor]
        ]:
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(
                    is_maintenance_mode=False,
                    failure_reason="Failed to stop ongoing mission",
                )
            )
            state_machine.logger.error(
                f"Failed to stop mission in StoppingDueToMaintenance. Message: {error_message.error_description}"
            )
            # TODO: see https://github.com/equinor/isar/issues/1047
            if mission_id == "":
                return ReturningHome.transition_to_existing_mission()
            return Monitor.transition_with_existing_mission(mission_id)

        def _successful_stop_event_handler(
            successful_stop: EmptyMessage,
        ) -> Transition[Maintenance.Maintenance]:
            state_machine.publish_mission_aborted(
                mission_id, "Mission aborted, robot being sent to maintenance", True
            )
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(is_maintenance_mode=True)
            )
            return Maintenance.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[ErrorMessage](
                name="failed_stop_event",
                event=events.robot_service_events.mission_failed_to_stop,
                handler=_failed_stop_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="successful_stop_event",
                event=events.robot_service_events.mission_successfully_stopped,
                handler=_successful_stop_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.StoppingDueToMaintenance,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition(mission_id: str) -> Transition[StoppingDueToMaintenance]:
    def _transition(state_machine: "StateMachine") -> StoppingDueToMaintenance:
        return StoppingDueToMaintenance(state_machine, mission_id)

    return _transition



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.pausing_return_home import PausingReturnHome
from isar.state_machine.states.resuming_return_home import ResumingReturnHome
from isar.state_machine.states.return_home_paused import ReturnHomePaused
from isar.state_machine.states.stopping_paused_return_home import (
    StoppingPausedReturnHome,
)
from robot_interface.models.mission.mission import Mission


def test_transition_from_pausing_return_home_to_return_home_paused(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = PausingReturnHome(sync_state_machine)

    pausing_return_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        pausing_return_home_state.get_event_handler_by_name("successful_pause_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturnHomePaused


def test_resuming_paused_return_home(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = ReturnHomePaused(sync_state_machine)

    return_home_paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        return_home_paused_state.get_event_handler_by_name("resume_return_home_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ResumingReturnHome


def test_transition_from_paused_return_home_to_stopping_paused_return_home_mission(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(90.0)
    sync_state_machine.current_state = ReturnHomePaused(sync_state_machine)

    return_home_paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        return_home_paused_state.get_event_handler_by_name("start_mission_event")
    )

    assert event_handler is not None

    example_mission: Mission = Mission(name="Dummy misson", tasks=[])

    transition = event_handler.handler(example_mission)

    sync_state_machine.current_state = transition(sync_state_machine)

    assert sync_state_machine.events.api_requests.start_mission.response.has_event()
    assert type(sync_state_machine.current_state) is StoppingPausedReturnHome


def test_stop_request_with_wrong_id_in_paused(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Paused(sync_state_machine, "mission_id")

    paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        paused_state.get_event_handler_by_name("stop_mission_event")
    )

    assert event_handler is not None

    transition = event_handler.handler("wrong_test_id")

    assert transition is None
    assert sync_state_machine.events.api_requests.stop_mission.response.has_event()



from threading import Thread

from isar.modules import ApplicationContainer
from isar.robot.robot import Robot
from isar.state_machine.state_machine import StateMachine, main
from isar.storage.uploader import Uploader


class StateMachineThreadMock(object):
    def __init__(self, container: ApplicationContainer) -> None:
        self.state_machine: StateMachine = container.state_machine()
        self._thread: Thread = Thread(target=main, args=[self.state_machine])

    def start(self) -> None:
        self._thread.start()

    def join(self) -> None:
        self.state_machine.terminate()
        self._thread.join()


class UploaderThreadMock(object):
    def __init__(self, container: ApplicationContainer) -> None:
        self.uploader: Uploader = container.uploader()
        self._thread: Thread = Thread(target=self.uploader.run)

    def start(self) -> None:
        self._thread.start()

    def join(self) -> None:
        self.uploader.stop()
        self._thread.join()


class RobotServiceThreadMock(object):
    def __init__(self, robot_service: Robot) -> None:
        self.robot_service: Robot = robot_service

    def start(self) -> None:
        self._thread: Thread = Thread(target=self.robot_service.run)
        self._thread.start()

    def join(self) -> None:
        self.robot_service.stop()
        self._thread.join()



import time
from collections import deque
from typing import Optional, cast

from pytest_mock import MockerFixture

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.modules import ApplicationContainer
from isar.robot.robot_status import RobotStatusThread
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.await_next_mission import AwaitNextMission
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.resuming import Resuming
from isar.state_machine.states.resuming_return_home import ResumingReturnHome
from isar.state_machine.states.return_home_paused import ReturnHomePaused
from isar.state_machine.states.unknown_status import UnknownStatus
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus
from tests.test_mocks.robot_interface import StubRobot
from tests.test_mocks.state_machine_mocks import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
    UploaderThreadMock,
)
from tests.test_mocks.task import StubTask


def test_state_machine_with_successful_mission_stop(
    container: ApplicationContainer,
    robot_service_thread: RobotServiceThreadMock,
    state_machine_thread: StateMachineThreadMock,
    uploader_thread: UploaderThreadMock,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    mocker.patch.object(
        StubRobot, "mission_status", return_value=MissionStatus.InProgress
    )

    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    mocker.patch.object(settings, "RETURN_HOME_DELAY", 15)

    mission: Mission = Mission(
        name="Dummy misson", tasks=[StubTask.take_image() for _ in range(0, 20)]
    )

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    state_machine_thread.start()
    robot_service_thread.start()
    uploader_thread.start()
    time.sleep(1)
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(0.5)
    scheduling_utilities.stop_mission(mission_id=mission.id)
    time.sleep(3)  # Allow enough time to stop the mission

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.Stopping,
            States.AwaitNextMission,
        ]
    )


def test_transition_from_resuming_to_paused(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Resuming(sync_state_machine, "mission_id")

    resuming_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        resuming_state.get_event_handler_by_name("failed_resume_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(
            error_reason=ErrorReason.RobotUnknownErrorException, error_description=""
        )
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Paused


def test_unknown_status_transitions_to_await_next_mission_if_it_was_already_available(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_status.trigger_event(RobotStatus.Available)
    # Make sure that we have not changed robot status
    sync_state_machine.events.robot_service_events.robot_status_changed.consume_event()

    sync_state_machine.current_state = UnknownStatus(sync_state_machine)

    unknown_status_state: State = cast(State, sync_state_machine.current_state)

    event_handler: Optional[EventHandlerMapping] = (
        unknown_status_state.get_event_handler_by_name("robot_status_event")
    )
    assert event_handler is not None

    transition = event_handler.handler(event_handler.event)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is AwaitNextMission


def test_transition_from_resuming_return_home_to_await_next_mission(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = ResumingReturnHome(sync_state_machine)

    resuming_return_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        resuming_return_home_state.get_event_handler_by_name("failed_resume_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(
            error_reason=ErrorReason.RobotUnknownErrorException, error_description=""
        )
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturnHomePaused



import logging
from queue import Queue
from threading import Event as ThreadEvent
from threading import Thread
from typing import Callable, List, Optional, Tuple

from isar.config.settings import settings
from isar.models.events import (
    EmptyMessage,
    Event,
    Events,
    RobotServiceEvents,
    SharedState,
    StateMachineEvents,
)
from isar.robot.robot_battery import RobotBatteryThread
from isar.robot.robot_monitor_mission import RobotMonitorMissionThread
from isar.robot.robot_pause_mission import RobotPauseMissionThread
from isar.robot.robot_resume_mission import RobotResumeMissionThread
from isar.robot.robot_start_mission import RobotStartMissionThread
from isar.robot.robot_status import RobotStatusThread
from isar.robot.robot_stop_mission import RobotStopMissionThread
from isar.robot.robot_upload_inspection import RobotUploadInspectionThread
from isar.services.utilities.mqtt_utilities import publish_mission_status
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus
from robot_interface.models.mission.task import TASKS
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface


class Robot(object):
    def __init__(
        self,
        events: Events,
        robot: RobotInterface,
        shared_state: SharedState,
        mqtt_publisher: MqttClientInterface,
    ) -> None:
        self.logger = logging.getLogger("robot")
        self.state_machine_events: StateMachineEvents = events.state_machine_events
        self.robot_service_events: RobotServiceEvents = events.robot_service_events
        self.mqtt_publisher: MqttClientInterface = mqtt_publisher
        self.upload_queue: Queue = events.upload_queue
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.start_mission_thread: Optional[RobotStartMissionThread] = None
        self.robot_battery_thread: Optional[RobotBatteryThread] = None
        self.robot_status_thread: Optional[RobotStatusThread] = None
        self.monitor_mission_thread: Optional[RobotMonitorMissionThread] = None
        self.stop_mission_thread: Optional[RobotStopMissionThread] = None
        self.pause_mission_thread: Optional[RobotPauseMissionThread] = None
        self.resume_mission_thread: Optional[RobotResumeMissionThread] = None
        self.upload_inspection_threads: List[RobotUploadInspectionThread] = []
        self.signal_thread_quitting: ThreadEvent = ThreadEvent()
        self.signal_mission_stopped: ThreadEvent = ThreadEvent()
        self.inspection_callback_thread: Optional[Thread] = None

    def stop(self) -> None:
        self.signal_thread_quitting.set()
        if self.robot_status_thread is not None and self.robot_status_thread.is_alive():
            self.robot_status_thread.join()
        if (
            self.robot_battery_thread is not None
            and self.robot_battery_thread.is_alive()
        ):
            self.robot_battery_thread.join()
        if (
            self.monitor_mission_thread is not None
            and self.monitor_mission_thread.is_alive()
        ):
            self.monitor_mission_thread.join()
        if (
            self.start_mission_thread is not None
            and self.start_mission_thread.is_alive()
        ):
            self.start_mission_thread.join()
        if self.stop_mission_thread is not None and self.stop_mission_thread.is_alive():
            self.stop_mission_thread.join()
        for thread in self.upload_inspection_threads:
            if thread.is_alive():
                thread.join()
        self.upload_inspection_threads = []
        self.robot_status_thread = None
        self.robot_battery_thread = None
        self.start_mission_thread = None
        self.monitor_mission_thread = None

    def _start_mission_done_handler(self) -> None:
        if (
            self.start_mission_thread is not None
            and not self.start_mission_thread.is_alive()
        ):
            self.start_mission_thread.join()
            mission = self.start_mission_thread.mission
            error_message = self.start_mission_thread.error_message
            self.start_mission_thread = None

            if (
                error_message
                and error_message.error_reason == ErrorReason.RobotAlreadyHomeException
            ):
                self.robot_service_events.robot_already_home.trigger_event(
                    EmptyMessage()
                )
                return
            elif error_message:
                mission.status = MissionStatus.Failed
                error_message.error_description = (
                    f"Failed to initiate due to: {error_message.error_description}"
                )
                mission.error_message = error_message
                publish_mission_status(
                    self.mqtt_publisher, mission.id, mission.status, error_message
                )
                self.robot_service_events.mission_failed.trigger_event(error_message)
                return

            self.logger.info("Received confirmation that mission has started")

            if (
                self.monitor_mission_thread is not None
                and self.monitor_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Attempted to start mission while monitoring an old mission."
                )
                self.monitor_mission_thread.join()

            self.signal_mission_stopped.clear()
            self.monitor_mission_thread = RobotMonitorMissionThread(
                lambda task: self.robot_service_events.request_inspection_upload.trigger_event(
                    (task, mission)
                ),
                self.robot,
                self.mqtt_publisher,
                self.signal_thread_quitting,
                self.signal_mission_stopped,
                mission,
            )
            self.monitor_mission_thread.start()

    def _stop_mission_done_handler(self) -> None:
        if (
            self.stop_mission_thread is not None
            and not self.stop_mission_thread.is_alive()
        ):
            self.stop_mission_thread.join()
            error_message = self.stop_mission_thread.error_message

            if error_message:
                self.robot_service_events.mission_failed_to_stop.trigger_event(
                    error_message
                )
                self.stop_mission_thread = None
            else:
                if (
                    self.monitor_mission_thread is not None
                    and self.monitor_mission_thread.is_alive()
                ):
                    self.signal_mission_stopped.set()
                    return

                self.stop_mission_thread = None
                # The mission status will already be reported on MQTT, the state machine does not need the event
                self.robot_service_events.mission_succeeded.clear_event()
                self.robot_service_events.mission_failed.clear_event()
                self.robot_service_events.mission_successfully_stopped.trigger_event(
                    EmptyMessage()
                )

    def _pause_mission_done_handler(self) -> None:
        if (
            self.pause_mission_thread is not None
            and not self.pause_mission_thread.is_alive()
        ):
            self.pause_mission_thread.join()
            error_message = self.pause_mission_thread.error_message
            self.pause_mission_thread = None

            if error_message:
                self.robot_service_events.mission_failed_to_pause.trigger_event(
                    error_message
                )
            else:
                self.robot_service_events.mission_successfully_paused.trigger_event(
                    EmptyMessage()
                )

    def _start_mission_event_handler(self, event: Event[Mission]) -> None:
        start_mission = event.consume_event()
        if start_mission is not None:
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Attempted to start mission while another mission was starting."
                )
                self.start_mission_thread.join()

            start_mission.status = MissionStatus.NotStarted
            publish_mission_status(
                self.mqtt_publisher, start_mission.id, start_mission.status, None
            )
            self.start_mission_thread = RobotStartMissionThread(
                self.robot,
                self.signal_thread_quitting,
                start_mission,
            )
            self.start_mission_thread.start()

    def _stop_mission_request_handler(self, event: Event[EmptyMessage]) -> None:
        if event.has_event():
            if (
                self.stop_mission_thread is not None
                and self.stop_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Received stop mission event while trying to stop a mission. Aborting stop attempt."
                )
                return
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                return
            event.consume_event()
            self.stop_mission_thread = RobotStopMissionThread(
                self.robot, self.signal_thread_quitting
            )
            self.stop_mission_thread.start()

    def _pause_mission_request_handler(self, event: Event[EmptyMessage]) -> None:
        if event.has_event():
            if (
                self.pause_mission_thread is not None
                and self.pause_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Received pause mission event while trying to pause a mission. Aborting pause attempt."
                )
                return
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                return
            event.consume_event()
            self.pause_mission_thread = RobotPauseMissionThread(
                self.robot, self.signal_thread_quitting
            )
            self.pause_mission_thread.start()

    def _resume_mission_request_handler(self, event: Event[EmptyMessage]) -> None:
        if event.consume_event():
            if (
                self.resume_mission_thread is not None
                and self.resume_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Received resume mission event while trying to resume a mission. Aborting resume attempt."
                )
                return
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                error_description = "Received resume mission event while trying to start a mission. Aborting resume attempt."
                error_message = ErrorMessage(
                    error_reason=ErrorReason.RobotStillStartingMissionException,
                    error_description=error_description,
                )
                self.robot_service_events.mission_failed_to_resume.trigger_event(
                    error_message
                )
                return
            self.resume_mission_thread = RobotResumeMissionThread(
                self.robot, self.signal_thread_quitting
            )
            self.resume_mission_thread.start()

    def _resume_mission_done_handler(self) -> None:
        if (
            self.resume_mission_thread is not None
            and not self.resume_mission_thread.is_alive()
        ):
            self.resume_mission_thread.join()
            error_message = self.resume_mission_thread.error_message
            self.resume_mission_thread = None

            if error_message:
                self.robot_service_events.mission_failed_to_resume.trigger_event(
                    error_message
                )
            else:
                self.robot_service_events.mission_successfully_resumed.trigger_event(
                    EmptyMessage()
                )

    def _upload_inspection_event_handler(
        self, event: Event[Tuple[TASKS, Mission]]
    ) -> None:
        upload_request = event.consume_event()
        if upload_request:

            upload_inspection_thread = RobotUploadInspectionThread(
                self.upload_queue, self.robot, upload_request[0], upload_request[1]
            )
            self.upload_inspection_threads.append(upload_inspection_thread)
            upload_inspection_thread.start()

    def _upload_inspection_done_handler(self) -> None:
        if len(self.upload_inspection_threads) > 0:

            def _join_threads(thread: RobotUploadInspectionThread) -> bool:
                if not thread.is_alive():
                    thread.join()
                    return True
                return False

            self.upload_inspection_threads[:] = [
                thread
                for thread in self.upload_inspection_threads
                if _join_threads(thread)
            ]

    def register_and_monitor_inspection_callback(
        self,
        callback_function: Callable,
    ) -> None:
        self.inspection_callback_function = callback_function

        self.inspection_callback_thread = self.robot.register_inspection_callback(
            callback_function
        )
        if self.inspection_callback_thread is not None:
            self.inspection_callback_thread.start()
            self.logger.info("Inspection callback thread started and will be monitored")

    def _monitor_inspection_callback_thread(self) -> None:
        if (
            self.inspection_callback_thread is not None
            and not self.inspection_callback_thread.is_alive()
        ):
            self.logger.warning("Inspection callback thread died - restarting")
            self.inspection_callback_thread.join()
            self.inspection_callback_thread.start()

    def _monitor_mission_done_handler(self) -> None:
        if (
            self.monitor_mission_thread is not None
            and not self.monitor_mission_thread.is_alive()
        ):
            mission_manually_cancelled = (
                self.monitor_mission_thread.signal_mission_stopped.is_set()
            )
            error_message = self.monitor_mission_thread.error_message
            if not mission_manually_cancelled:
                if error_message is not None:
                    self.robot_service_events.mission_failed.trigger_event(
                        error_message
                    )
                else:
                    self.robot_service_events.mission_succeeded.trigger_event(
                        EmptyMessage()
                    )
            self.monitor_mission_thread.join()
            self.monitor_mission_thread = None

    def run(self) -> None:
        self.robot_status_thread = RobotStatusThread(
            robot=self.robot,
            signal_thread_quitting=self.signal_thread_quitting,
            shared_state=self.shared_state,
            state_machine_events=self.state_machine_events,
            robot_service_events=self.robot_service_events,
        )
        self.robot_status_thread.start()

        self.robot_battery_thread = RobotBatteryThread(
            self.robot, self.signal_thread_quitting, self.shared_state
        )
        self.robot_battery_thread.start()

        try:
            while not self.signal_thread_quitting.wait(0):
                self._start_mission_event_handler(
                    self.state_machine_events.start_mission
                )

                self._pause_mission_request_handler(
                    self.state_machine_events.pause_mission
                )

                self._resume_mission_request_handler(
                    self.state_machine_events.resume_mission
                )

                self._stop_mission_request_handler(
                    self.state_machine_events.stop_mission
                )

                self._upload_inspection_event_handler(
                    self.robot_service_events.request_inspection_upload
                )

                self._start_mission_done_handler()

                self._stop_mission_done_handler()

                self._pause_mission_done_handler()

                self._upload_inspection_done_handler()

                self._resume_mission_done_handler()

                self._monitor_mission_done_handler()

                if settings.UPLOAD_INSPECTIONS_ASYNC:
                    self._monitor_inspection_callback_thread()
        except Exception as e:
            self.logger.error(f"Unhandled exception in robot service: {str(e)}")
        self.logger.info("Exiting robot service main thread")



import json
import logging
from collections import deque
from datetime import datetime, timezone
from threading import Event
from typing import Deque, Optional

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import State
from isar.models.events import Events, SharedState
from isar.models.status import IsarStatus
from isar.services.service_connections.persistent_memory import (
    NoSuchRobotException,
    create_persistent_robot_state,
    read_persistent_robot_state_is_maintenance_mode,
)
from isar.services.utilities.mqtt_utilities import publish_isar_status
from isar.state_machine.states.maintenance import Maintenance
from isar.state_machine.states.unknown_status import UnknownStatus
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import ReturnToHome
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.telemetry.payloads import (
    InterventionNeededPayload,
    MissionAbortedPayload,
)
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class StateMachine(object):
    """Handles state transitions for supervisory robot control."""

    def __init__(
        self,
        events: Events,
        shared_state: SharedState,
        robot: RobotInterface,
        mqtt_publisher: MqttClientInterface,
    ):
        """Initializes the state machine.

        Parameters
        ----------
        events : Events
            Events used for API and robot service communication.
        robot : RobotInterface
            Instance of robot interface.
        mqtt_publisher : MqttClientInterface
            Instance of MQTT client interface which has a publish function

        """
        self.logger = logging.getLogger("state_machine")

        self.events: Events = events
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.mqtt_publisher: Optional[MqttClientInterface] = mqtt_publisher

        self.signal_state_machine_to_stop: Event = Event()

        self.current_state: State = UnknownStatus(self)

        if settings.PERSISTENT_STORAGE_CONNECTION_STRING == "":
            self.logger.warning(
                "PERSISTENT_STORAGE_CONNECTION_STRING is not set. Restarting ISAR will forget the state, including maintenance mode. "
            )
        else:
            is_maintenance_mode = read_or_create_persistent_maintenance_mode()
            self.logger.info(
                f"Connected to robot status database and the maintenance mode was: {is_maintenance_mode}. "
            )
            if is_maintenance_mode:
                self.current_state = Maintenance(self)

        self.transitions_list: Deque[States] = deque(
            [], settings.STATE_TRANSITIONS_LOG_LENGTH
        )

    #################################################################################

    def print_transitions(self) -> None:
        state_transitions: str = ", ".join(
            [
                f"\n  {transition}" if (i + 1) % 10 == 0 else f"{transition}"
                for i, transition in enumerate(list(self.transitions_list))
            ]
        )
        self.logger.info("State transitions:\n  %s", state_transitions)
        self.transitions_list.clear()

    def run(self) -> None:
        """Runs the state machine loop."""
        try:
            while self.current_state is not None:
                self.update_state()
                self.current_state = self.current_state.run()
                if self.current_state is None:
                    self.logger.warning(
                        "Exiting state machine as current state is None"
                    )
        except Exception as e:
            self.logger.error(f"Unhandled exception in state machine: {str(e)}")

    def terminate(self) -> None:
        self.logger.info("Stopping state machine")
        self.signal_state_machine_to_stop.set()

    def battery_level_is_above_mission_start_threshold(self) -> bool:
        if not self.shared_state.robot_battery_level.check():
            self.logger.warning("Battery level is None")
            return False
        return (
            not self.shared_state.robot_battery_level.check()
            < settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
        )

    def update_state(self) -> None:
        """Updates the current state of the state machine."""
        self.shared_state.state.update(self.current_state.name)
        self.transitions_list.append(self.current_state.name)
        self.logger.info("State: %s", self.current_state.name)
        self.publish_status()

    def start_mission(self, mission: Mission) -> None:
        """Starts a scheduled mission."""
        self.events.state_machine_events.start_mission.trigger_event(mission)

    def start_return_home_mission(self) -> None:
        """Starts a return to home mission."""
        mission = Mission(
            tasks=[ReturnToHome()],
            name="Return Home",
        )
        self.events.state_machine_events.start_mission.trigger_event(mission)

    def publish_mission_aborted(
        self, current_mission_id: Optional[str], reason: str, can_be_continued: bool
    ) -> None:
        if not self.mqtt_publisher:
            return

        if current_mission_id is None:
            self.logger.warning(
                "Publishing mission aborted message with no ongoing mission."
            )

        payload: MissionAbortedPayload = MissionAbortedPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            mission_id=current_mission_id,
            reason=reason,
            can_be_continued=can_be_continued,
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_MISSION_ABORTED,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_intervention_needed(self, error_message: str) -> None:
        """Publishes the intervention needed message to the MQTT Broker"""
        if not self.mqtt_publisher:
            return

        payload: InterventionNeededPayload = InterventionNeededPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            reason=error_message,
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_INTERVENTION_NEEDED,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_status(self) -> None:
        if not self.mqtt_publisher:
            return

        publish_isar_status(self.mqtt_publisher, self._current_status())

    def _current_status(self) -> IsarStatus:
        if self.current_state.name == States.AwaitNextMission:
            return IsarStatus.Available
        elif self.current_state.name == States.ReturnHomePaused:
            return IsarStatus.ReturnHomePaused
        elif self.current_state.name == States.Paused:
            return IsarStatus.Paused
        elif self.current_state.name == States.Home:
            return IsarStatus.Home
        elif self.current_state.name == States.ReturningHome:
            return IsarStatus.ReturningHome
        elif self.current_state.name == States.Offline:
            return IsarStatus.Offline
        elif self.current_state.name == States.BlockedProtectiveStop:
            return IsarStatus.BlockedProtectiveStop
        elif self.current_state.name == States.InterventionNeeded:
            return IsarStatus.InterventionNeeded
        elif self.current_state.name == States.Recharging:
            return IsarStatus.Recharging
        elif self.current_state.name == States.Lockdown:
            return IsarStatus.Lockdown
        elif self.current_state.name == States.GoingToLockdown:
            return IsarStatus.GoingToLockdown
        elif self.current_state.name == States.GoingToRecharging:
            return IsarStatus.GoingToRecharging
        elif self.current_state.name == States.Maintenance:
            return IsarStatus.Maintenance
        elif self.current_state.name == States.Pausing:
            return IsarStatus.Pausing
        elif self.current_state.name == States.PausingReturnHome:
            return IsarStatus.PausingReturnHome
        elif self.current_state.name in [
            States.Stopping,
            States.StoppingDueToMaintenance,
            States.StoppingGoToLockdown,
            States.StoppingGoToRecharge,
            States.StoppingPausedMission,
            States.StoppingPausedReturnHome,
        ]:
            return IsarStatus.Stopping
        elif self.current_state.name == States.StoppingReturnHome:
            return IsarStatus.StoppingReturnHome
        else:
            return IsarStatus.Busy


def read_or_create_persistent_maintenance_mode() -> bool:
    try:
        is_maintenance_mode = read_persistent_robot_state_is_maintenance_mode(
            settings.PERSISTENT_STORAGE_CONNECTION_STRING, settings.ISAR_ID
        )
    except NoSuchRobotException:
        create_persistent_robot_state(
            settings.PERSISTENT_STORAGE_CONNECTION_STRING, settings.ISAR_ID
        )
        is_maintenance_mode = read_persistent_robot_state_is_maintenance_mode(
            settings.PERSISTENT_STORAGE_CONNECTION_STRING, settings.ISAR_ID
        )
    return is_maintenance_mode


def main(state_machine: StateMachine) -> None:
    """Starts a state machine instance."""
    state_machine.run()



import logging
import time
from abc import ABC
from copy import deepcopy
from dataclasses import dataclass
from threading import Event as ThreadEvent
from typing import TYPE_CHECKING, Any, Callable, Generic, List, Optional, TypeVar

from isar.config.settings import settings
from isar.models.events import Event
from isar.state_machine.states_enum import States

T = TypeVar("T")

T_state = TypeVar("T_state", bound="State", covariant=True)
Transition = Callable[["StateMachine"], T_state]


@dataclass
class EventHandlerMapping(Generic[T]):
    name: str
    event: Event[T]
    handler: Callable[[T], Optional[Transition]]
    should_not_consume: bool = False


@dataclass
class TimeoutHandlerMapping:
    name: str
    timeout_in_seconds: float
    handler: Callable[[], Optional[Transition]]


if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class State(ABC):
    def __init__(
        self,
        state_machine: "StateMachine",
        state_name: States,
        event_handler_mappings: List[EventHandlerMapping],
        timers: List[TimeoutHandlerMapping] = [],
    ) -> None:
        self.name = state_name
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.events = state_machine.events
        self.signal_state_machine_to_stop: ThreadEvent = (
            state_machine.signal_state_machine_to_stop
        )
        self.event_handler_mappings = event_handler_mappings
        self.timers = timers

    def get_event_handler_by_name(
        self, event_handler_name: str
    ) -> Optional[EventHandlerMapping]:
        filtered_handlers = list(
            filter(
                lambda mapping: mapping.name == event_handler_name,
                self.event_handler_mappings,
            )
        )
        return filtered_handlers[0] if len(filtered_handlers) > 0 else None

    def get_event_timer_by_name(
        self, event_timer_name: str
    ) -> Optional[TimeoutHandlerMapping]:
        filtered_timers = list(
            filter(
                lambda mapping: mapping.name == event_timer_name,
                self.timers,
            )
        )
        return filtered_timers[0] if len(filtered_timers) > 0 else None

    def run(self) -> Optional["State"]:
        should_exit_state: bool = False
        timers = deepcopy(self.timers)
        entered_time = time.time()
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info("Stopping state machine from %s state", self.name)
                break

            for timer in timers:
                if time.time() - entered_time > timer.timeout_in_seconds:
                    transition = timer.handler()
                    timers.remove(timer)
                    if transition is not None:
                        return transition(self.state_machine)

            if should_exit_state:
                break

            for handler_mapping in self.event_handler_mappings:
                event_value: Optional[Any]
                if handler_mapping.should_not_consume:
                    event_value = handler_mapping.event.check()
                else:
                    event_value = handler_mapping.event.consume_event()
                if event_value is not None:
                    transition = handler_mapping.handler(event_value)
                    if transition is not None or not handler_mapping.should_not_consume:
                        self.logger.debug(
                            f"Event '{handler_mapping.name}' triggered with input: {event_value}. "
                        )
                        if transition is not None:
                            self.logger.debug(
                                f"Transitioning from {self.name.name} to {transition.__annotations__['return'].__name__}"
                            )
                    if transition is not None:
                        return transition(self.state_machine)

            if should_exit_state:
                break
            time.sleep(settings.FSM_SLEEP_TIME)
        return None



import time
from datetime import datetime
from queue import Queue
from threading import Thread
from typing import Callable, List, Optional

from alitra import Frame, Orientation, Pose, Position

from isar.models.events import Event
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import (
    RobotCommunicationException,
    RobotMissionStatusException,
)
from robot_interface.models.inspection.inspection import (
    Image,
    ImageMetadata,
    Inspection,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import InspectionTask
from robot_interface.models.robots.media import MediaConfig, MediaConnectionType
from robot_interface.robot_interface import RobotInterface


class StubRobot(RobotInterface):
    def __init__(
        self,
        mission_status: MissionStatus = MissionStatus.Successful,
        task_status: TaskStatus = TaskStatus.Successful,
        stop: bool = True,
        pose: Pose = Pose(
            position=Position(x=0, y=0, z=0, frame=Frame("robot")),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
            frame=Frame("robot"),
        ),
        robot_status: RobotStatus = RobotStatus.Available,
        initiate_mission_delay: float = 0.0,
    ) -> None:
        self.mission_status_return_value: MissionStatus = mission_status
        self.task_status_return_value: TaskStatus = task_status
        self.stop_return_value: bool = stop
        self.robot_pose_return_value: Pose = pose
        self.robot_status_return_value: RobotStatus = robot_status
        self.initiate_mission_delay: float = initiate_mission_delay
        self.mission: Optional[Mission] = None

    def initiate_mission(self, mission: Mission) -> None:
        time.sleep(self.initiate_mission_delay)
        self.mission = mission
        return

    def task_status(self, task_id: str) -> TaskStatus:
        return self.task_status_return_value

    def mission_status(self, mission_id: str) -> MissionStatus:
        return self.mission_status_return_value

    def stop(self) -> None:
        return

    def pause(self) -> None:
        return

    def resume(self) -> None:
        return

    def get_inspection(self, task: InspectionTask) -> Inspection:
        return Image(
            metadata=stub_image_metadata(),
            id=task.inspection_id,
            data=b"Some binary image data",
        )

    def generate_media_config(self) -> MediaConfig:
        return MediaConfig(
            url="stubURL",
            token="stubToken",
            media_connection_type=MediaConnectionType.LiveKit,
        )

    def register_inspection_callback(
        self, callback_function: Callable[[Inspection, Mission], None]
    ) -> None:
        return

    def get_telemetry_publishers(
        self, queue: Queue, isar_id: str, robot_name: str
    ) -> List[Thread]:
        return []

    def robot_status(self) -> RobotStatus:
        return self.robot_status_return_value

    def get_battery_level(self) -> float:
        return 80.0


def stub_image_metadata() -> ImageMetadata:
    return ImageMetadata(
        start_time=datetime.now(),
        robot_pose=Pose(
            Position(0, 0, 0, Frame("robot")),
            Orientation(0, 0, 0, 1, Frame("robot")),
            Frame("robot"),
        ),
        target_position=Position(0, 0, 0, Frame("robot")),
        file_type="jpg",
    )


class StubRobotInitiateMissionRaisesException(StubRobot):
    def __init__(self) -> None:
        super().__init__()

    def initiate_mission(self, mission: Mission) -> None:
        raise RobotCommunicationException("Testing Initate Mission Exception")


class StubRobotMissionStatusRaisesException(StubRobot):
    def __init__(self) -> None:
        super().__init__()

    def mission_status(self, mission_id: str) -> MissionStatus:
        raise RobotMissionStatusException("Testing Mission Status Exception")


class StubRobotOfflineToAvailableTest(StubRobot):
    def __init__(self, current_state: Event) -> None:
        self.entered_offline = False
        self.current_state = current_state

    def robot_status(self) -> RobotStatus:
        new_state = self.current_state.check()
        if new_state is None:
            raise RobotCommunicationException("Could not read state machine state")
        if new_state == "offline":
            self.entered_offline = True
            return RobotStatus.Available

        if not self.entered_offline:
            return RobotStatus.Offline

        return RobotStatus.Available


class StubRobotBlockedProtectiveStopToHomeTest(StubRobot):
    def __init__(self, current_state_event: Event[States]) -> None:
        self.entered_blocked_p_stop = False
        self.current_state_event = current_state_event

    def robot_status(self) -> RobotStatus:
        current_state: States = self.current_state_event.check()
        if current_state is None:
            raise RobotCommunicationException("Could not read state machine state")
        if current_state == States.BlockedProtectiveStop:
            self.entered_blocked_p_stop = True
            return RobotStatus.Home
        if not self.entered_blocked_p_stop:
            return RobotStatus.BlockedProtectiveStop
        return RobotStatus.Home


class StubRobotOfflineToHomeTest(StubRobot):
    def __init__(self, current_state: Event) -> None:
        self.entered_offline = False
        self.current_state = current_state

    def robot_status(self) -> RobotStatus:
        current_state = self.current_state.check()
        if current_state is None:
            raise RobotCommunicationException("Could not read state machine state")
        if current_state == "offline":
            self.entered_offline = True
            return RobotStatus.Home
        if not self.entered_offline:
            return RobotStatus.Offline
        return RobotStatus.Home


class StubRobotRobotStatusBusyIfNotHomeOrUnknownStatus(StubRobot):
    def __init__(
        self,
        current_state: Event,
        initiate_mission_delay: float = 0.0,
    ) -> None:
        super().__init__()
        self.current_state = current_state
        self.initiate_mission_delay: float = initiate_mission_delay
        self.return_home_mission_just_finished_successfully = False

    def task_status(self, task_id: str) -> TaskStatus:
        if self.mission._is_return_to_home_mission():
            if self.task_status_return_value == TaskStatus.Successful:
                self.return_home_mission_just_finished_successfully = True
        return self.task_status_return_value

    def robot_status(self) -> RobotStatus:
        current_state = self.current_state.check()
        if current_state is None:
            raise RobotCommunicationException("Could not read state machine state")
        if current_state == States.Home:
            return RobotStatus.Home
        elif current_state == States.UnknownStatus:
            return RobotStatus.Home
        elif self.return_home_mission_just_finished_successfully:
            return RobotStatus.Home

        return RobotStatus.Busy



import logging

from robot_interface.models.robots.media import MediaConfig
from robot_interface.robot_interface import RobotInterface


class RobotUtilities:
    """
    Contains utility functions for getting robot information from the API.
    """

    def __init__(
        self,
        robot: RobotInterface,
    ):
        self.robot: RobotInterface = robot
        self.logger = logging.getLogger("api")

    def generate_media_config(self) -> MediaConfig:
        return self.robot.generate_media_config()



import logging
from pathlib import Path

from isar.config.settings import settings
from isar.storage.storage_interface import (
    LocalStoragePath,
    StorageException,
    StorageInterface,
    StoragePaths,
)
from isar.storage.utilities import construct_metadata_file, construct_paths
from robot_interface.models.inspection.inspection import InspectionBlob
from robot_interface.models.mission.mission import Mission


class LocalStorage(StorageInterface):
    def __init__(self) -> None:
        self.root_folder: Path = Path(settings.LOCAL_STORAGE_PATH)
        self.logger = logging.getLogger("uploader")

    def store(
        self, inspection: InspectionBlob, mission: Mission
    ) -> StoragePaths[LocalStoragePath]:
        if inspection.data is None:
            raise StorageException("Nothing to store. The inspection data is empty")

        local_filename, local_metadata_filename = construct_paths(
            inspection=inspection, mission=mission
        )

        data_path: Path = self.root_folder.joinpath(local_filename)
        metadata_path: Path = self.root_folder.joinpath(local_metadata_filename)

        data_path.parent.mkdir(parents=True, exist_ok=True)

        metadata_bytes: bytes = construct_metadata_file(
            inspection=inspection, mission=mission, filename=local_filename.name
        )
        try:
            with (
                open(data_path, "wb") as file,
                open(metadata_path, "wb") as metadata_file,
            ):
                file.write(inspection.data)
                metadata_file.write(metadata_bytes)
        except IOError as e:
            self.logger.warning(
                f"Failed open/write for one of the following files: \n"
                f"{data_path}\n{metadata_path}"
            )
            raise StorageException from e
        except Exception as e:
            self.logger.error(
                "An unexpected error occurred while writing to local storage"
            )
            raise StorageException from e
        return StoragePaths(
            data_path=LocalStoragePath(file_path=data_path),
            metadata_path=LocalStoragePath(file_path=metadata_path),
        )



from enum import Enum
from typing import Literal, Optional, Type, Union

from alitra import Pose, Position
from pydantic import BaseModel, Field

from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.inspection.inspection import (
    Audio,
    CO2Measurement,
    Image,
    Inspection,
    ThermalImage,
    ThermalVideo,
    Video,
)
from robot_interface.models.mission.status import TaskStatus
from robot_interface.utilities.uuid_string_factory import uuid4_string


class TaskTypes(str, Enum):
    ReturnToHome = "return_to_home"
    TakeImage = "take_image"
    TakeThermalImage = "take_thermal_image"
    TakeVideo = "take_video"
    TakeThermalVideo = "take_thermal_video"
    TakeCO2Measurement = "take_co2_measurement"
    RecordAudio = "record_audio"


class ZoomDescription(BaseModel):
    objectWidth: float
    objectHeight: float


class Task(BaseModel):
    status: TaskStatus = Field(default=TaskStatus.NotStarted)
    error_message: Optional[ErrorMessage] = Field(default=None)
    tag_id: Optional[str] = Field(default=None)
    id: str = Field(default_factory=uuid4_string, frozen=True)


class InspectionTask(Task):
    """
    Base class for all inspection tasks which produce results to be uploaded.
    """

    inspection_id: str = Field(default_factory=uuid4_string, frozen=True)
    robot_pose: Pose = Field(default=None, init=True)
    inspection_description: Optional[str] = Field(default=None)
    zoom: Optional[ZoomDescription] = Field(default=None)

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Inspection


class ReturnToHome(Task):
    """
    Task which cases the robot to return home
    """

    type: Literal[TaskTypes.ReturnToHome] = TaskTypes.ReturnToHome


class TakeImage(InspectionTask):
    """
    Task which causes the robot to take an image towards the given target.
    """

    target: Position = Field(default=None)
    type: Literal[TaskTypes.TakeImage] = TaskTypes.TakeImage

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Image


class TakeThermalImage(InspectionTask):
    """
    Task which causes the robot to take a thermal image towards the given target.
    """

    target: Position = Field(default=None)
    type: Literal[TaskTypes.TakeThermalImage] = TaskTypes.TakeThermalImage

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return ThermalImage


class TakeVideo(InspectionTask):
    """
    Task which causes the robot to take a video towards the given target.

    Duration of video is given in seconds.
    """

    target: Position = Field(default=None)
    duration: float = Field(default=None)
    type: Literal[TaskTypes.TakeVideo] = TaskTypes.TakeVideo

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Video


class TakeThermalVideo(InspectionTask):
    """
    Task which causes the robot to record thermal video towards the given target

    Duration of video is given in seconds.
    """

    target: Position = Field(default=None)
    duration: float = Field(default=None)
    type: Literal[TaskTypes.TakeThermalVideo] = TaskTypes.TakeThermalVideo

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return ThermalVideo


class RecordAudio(InspectionTask):
    """
    Task which causes the robot to record a video at its position, facing the target.

    Duration of audio is given in seconds.
    """

    target: Position = Field(default=None)
    duration: float = Field(default=None)
    type: Literal[TaskTypes.RecordAudio] = TaskTypes.RecordAudio

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Audio


class TakeCO2Measurement(InspectionTask):
    """
    Task which causes the robot to take a CO2 measurement at its position.
    """

    type: Literal[TaskTypes.TakeCO2Measurement] = TaskTypes.TakeCO2Measurement

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return CO2Measurement


TASKS = Union[
    ReturnToHome,
    TakeImage,
    TakeThermalImage,
    TakeVideo,
    TakeThermalVideo,
    TakeCO2Measurement,
    RecordAudio,
]
